package WWW::OrangeHRM::Client;
use strict;
use warnings;
use utf8;
use version 0.77; our $VERSION = qv("v0.1.1");

use WWW::Mechanize;
use HTML::TreeBuilder::XPath;

use Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(debug_http dump_state fatal_error log_in log_out time_sheet
    time_sheet_show normalize_time time_sheet_set);

=encoding utf8

=head1 NAME

WWW::OrangeHRM::Client - Client for OrangeHRM

=head1 DESCRIPTION

This module implements client for
L<OrangeHRM|http://en.wikipedia.org/wiki/OrangeHRM> information system. It has
been developed against Red Hat instance, so I cannot guarantee it works with
other instances. Author and this code have no business or personal relation to
OrangeHRM Inc.

=head1 SYNOPSIS

    use WWW::OrangeHRM::Client;
    my %configuration = (
        url => 'https://redhat.orangehrm.com/',
        username => 'foo',
        password => 'bar'
    );
    my $debug = 0;
    my $automaton = WWW::Mechanize->new();
    
    # Log-in
    if (!WWW::OrangeHRM::Client::log_in($automaton, \%configuration)) {
        WWW::OrangeHRM::Client::fatal_error($automaton, $debug,
            'Could not log in!');
    }
    print "Logged in.\n";
    
    if (!WWW::OrangeHRM::Client::time_sheet($automaton)) {
        WWW::OrangeHRM::Client::fatal_error($automaton, $debug,
            'Could not get time sheet!');
    }
    
    WWW::OrangeHRM::Client::time_sheet_show($automaton);
    
    # Log-out
    if (!WWW::OrangeHRM::Client::log_out($automaton)) {
        WWW::OrangeHRM::Client::fatal_error($automaton, $debug,
            'Could not log out!');
    }
    print "Logged out.\n";

=cut

my $last_request;

BEGIN {
    $last_request = undef;
}


=head1 API

$automaton is C<WWW::Mechanize> object.

=head2 %configuration

    my %configuration = (
        url => 'https://redhat.orangehrm.com/',
        username => 'foo@redhat.com',
        password => undef
    );

=cut 

=head2 debug_http($automaton)

Call this after creating $automaton if you want to see HTTP request in
dump_state() or fatal_errro() dump.

=cut

sub debug_http {
    my $automaton = shift;
    $automaton->add_handler("request_send", sub {
            $last_request = shift->dump(maxlength => 0);
            return;
        });
    1;
}


=head2 dump_state($automaton)

Print details about last request and response.

=cut

sub dump_state {
    my $automaton = shift;

    if (defined $last_request) {
        print "===BEGIN HTTP REQUEST===\n";
        print $last_request;
        print "===END HTTP REQUEST===\n";
    }

    print 'Current URI: ', $automaton->uri, "\n";

    print "===BEGIN RESPONSE HEADERS===\n";
    $automaton->dump_headers();
    print "===END RESPONSE HEADERS===\n";

    print "===BEGIN RESPONSE BODY===\n";
    print $automaton->response->decoded_content, "\n";
    print "===END RESPONSE BODY===\n";
    
    print "===BEGIN TEXT DUMP===\n";
    $automaton->dump_text();
    print "===END TEXT DUMP===\n";
    
    print "===BEGIN LINK DUMP===\n";
    $automaton->dump_links();
    print "===END LINK DUMP===\n";
    
    print "===BEGIN FORM DUMP===\n";
    $automaton->dump_forms();
    print "===END FORM DUMP===\n";

    return 1;
}


=head2 fatal_error($automaton, $debug, $message)

Call this function, if you encounter fatal error.

$message will be printed. Automaton state will be dumped if $debug is true.

This function does not return, it exits program with an error exit code.

=cut

sub fatal_error {
    my ($automaton, $debug, $message) = @_;

    print $message, "\n";
    if ($debug) {
        dump_state($automaton);
        print $message, "\n";
    }
    
    exit 1;
}


=head2 get_text_status($automaton)

Return text in *[@id='messageBalloonContainer'] element of current HTML
document or undef.

=cut

sub get_text_status {
    my $automaton = shift;
    my $document = $automaton->response->decoded_content;
    my $parser = HTML::TreeBuilder::XPath->new()->parse($document)
        or return undef;
    return $parser->findvalue(q{//*[@id="messageBalloonContainer"]});
}


=head2 log_in($automaton, \%configuration)

Log into system and user as defined in %configuration.

Return false in case of failure, true in case of success.

=cut

sub log_in {
    my ($automaton, $configuration) = @_;

    print "Getting log-in page for <", $configuration->{url}, ">\n";
    $automaton->get($configuration->{url});

    print "Logging in as `", $configuration->{username}, "'\n";
    $automaton->submit_form(
        form_name => 'frmLogin',
        fields  => {
            'login[txtUserName]' => $configuration->{username},
            'login[txtUserPassword]' => $configuration->{password}
        },
    );
    if ($automaton->uri !~ m{/leave/redHatViewLeaveList\z}) {
        return 0;
    }
    return 1;
}


=head2 log_out($automaton)

Log out the system.

Return false in case of failure, true in case of success.

=cut

sub log_out {
    my $automaton = shift;

    print "Logging out\n";
    $automaton->get('/index.php/auth/logOut');

    if ($automaton->uri !~ m{/index.php/\z}) {
        return 0;
    }

    return 1;
}


=head2 time_sheet($automaton)

Retrieve current time sheet.

Return false in case of failure, true in case of success.

=cut

sub time_sheet {
    my $automaton = shift;
    my $path = '/index.php/time/editTimesheet/show/mine';

    print "Getting time sheet\n";
    $automaton->get($path);

    if ($automaton->uri !~ m{$path\z}) {
        return 0;
    }

    print "Time sheet retrieved successfully.\n";
    #dump_state($automaton);
    return 1;
}


=head2 time_sheet_show($automaton)

Show current time sheet.

Return false in case of failure, true in case of success.

=cut

sub time_sheet_show {
    my $automaton = shift;
    my $path = '/index.php/time/editTimesheet/show/mine';

    if ($automaton->uri !~ m{$path\z}) {
        return 0;
    }

    #dump_state($automaton);

    my $document = $automaton->response->decoded_content;
    my $parser = HTML::TreeBuilder::XPath->new()->parse($document)
        or return 0;
    my $form = $automaton->form_id('editTimesheet') or return 0;

    printf "Time sheet for %s-%s:\n",
        $form->value('editTimesheet[yearChoices]') // '<unknown year>',
        $form->value('editTimesheet[monthChoices]') // '<uknown month>';

    print "Day\tDayType\tWhere\tFrom\tTo\tBreak\tDoctor\tTotal\tComment\n";
    for (map {$_->name} ($form->inputs)) {
        next if ! defined;
        if (/\Aoriginal_record_ids\[([0,1,2,3,4,5,6,7,8,9]+)\]\z/) {
            my $index = $1;
            print $index + 1;

            my $class = $parser->findvalue(
                q{//tr[td/input/@id="record_id_}. $index . q{"]/@class}
            ) // '';
            if ($class eq 'days_off') {
                $class = 'weekend';
            } elsif ($class eq 'full_day_holiday') {
                $class = 'holiday';
            } elsif ($class eq 'leave_day') {
                # ??? What is it?
                $class = 'sickday';
            } elsif ($class eq 'leave_date') {
                # This is both sickday and leave.
                # TODO: Discriminate on table[@class=~'leave_info_table']
                $class = 'leave'
            } elsif ($class eq 'half_day_leave') {
                $class = 'leave/2'
            }
            print "\t$class";

            if ($form->value("classification_id\[$index\]") == 2) {
                print "\ttrip";
            } else {
                print "\twork";
            }

            for (qw(in_time out_time break_duration medical_checkup_duration
                row_total_duration comment)) {
                print "\t", $form->value("$_\[$index\]") // '';
            }

            print "\n";
        }
    }


    printf "Total leave: %s\n",
        $parser->findvalue(q{//*[@id="totalLeave"]/@value}) // '';
    printf "Total time entered: %s\n",
        $parser->findvalue(
            q{//th[text()="Total Time Entered"]/following-sibling::td[1]}
        ) // '';
    printf "Total time: %s\n",
        $parser->findvalue(q{//*[@id="total"]/@value}) // '';
    printf "Total time expected: %s\n",
        $parser->findvalue(q{//*[@id="expected"]/@value}) // '';
    return 1;
}


=head2 normalize_time($time)

Return $time value normalized to HH:MM format.

Return undef if $time is invalid.

This function can differ from web interface.

=cut

sub normalize_time {
    $_ = shift;
    if (!defined) {
        return undef;
    }

    my ($hours, $minutes);
    if (($hours) = (m/\A([0123456789]*)\z/)) {
        $minutes = '00';
    } elsif (!(  ($hours, $minutes) =
            (m/\A([0123456789]*):([0123456789]*)\z/)
        )) {
        print STDERR "Invalid time value `$_'\n";
        return undef;
    }
    
    for ($hours, $minutes) {
        if (length($_) == 0) {
            $_ = '00';
        } elsif (length($_) == 1) {
            $_ = '0' . $_;
        }
    }

    return $hours . ':' . $minutes;
}


=head2 time_sheet_set($automaton, $day, $from, $to, $break, $doctor,
$comment, $trip)

Fill a day into time sheet as if you worked from $from to $to with $break
and $doctor hours. The format is HH:MM. $comment is free text, $trip is true
for bussines trip, otherwise standard work.

Return false in case of failure, true in case of success.

=cut

sub time_sheet_set {
    my ($automaton, $day, $from, $to, $break, $doctor, $comment, $trip) = @_;
    my $index = $day - 1;
    my $classification = ($trip ? 2 : 1);
    $doctor = normalize_time($doctor);  # Server refuses H:MM $doctor value

    print "Filling day #", $day, " into time sheet\n";

    my $form = $automaton->form_id('editTimesheet') or return 0;
    $automaton->set_fields(
        "classification_id[$index]" => $classification,
        "in_time[$index]" => $from,
        "out_time[$index]" => $to,
        "break_duration[$index]" => $break,
        "medical_checkup_duration[$index]" => $doctor,
        "comment[$index]" => $comment,
        # This is set by Firefox magically. Probably a JS voodoo.
        #"row_total_duration[$index]" not necessary, server computes it.
        "editTimesheet[txtAction]" => 'save'
    );
    # Server expects disabled values in the form, otherwise it reports failure
    # despite saving the data. Server disables some inputs on full-day
    # leave.
    for (grep {$_->disabled} ($form->inputs)) {
        $_->disabled(0);
    }
    #print "DEBUG: ", $form->dump(), "\n";
    # WWW::Mechanize cannot find the 'Save' button by value for uknown reason.
    my $submit_button = $form->find_input('#btnSave') or return 0;
    $automaton->click_button(input => $submit_button);

    if ($automaton->uri !~ m{/index.php/time/editTimesheet\z}) {
        return 0;
    }
    my $message = get_text_status($automaton);
    if (! defined $message) {
        return 0;
    }
    if (! ($message eq 'Timesheet Successfully Saved') ) {
        print "Saving time sheet failed: $message\n";
        return 0;
    }

    print "Day saved successfully.\n";
    return 1;
}

1;
__END__

=head1 COPYRIGHT

Copyright © 2012  Petr Písař <ppisar@redhat.com>.

=head1 LICENSE

This is free software.  You may redistribute copies of it under the terms of
the GNU General Public License L<http://www.gnu.org/licenses/gpl.html>.
There is NO WARRANTY, to the extent permitted by law.

=cut
